'''
	attridict.py:
		Alvin Shaita <alvinshaita@gmail.com>
'''

__author__	= "Alvin Shaita"
__email__	= "alvinshaita@gmail.com"


class AttriDictAttributes:
	__version__ = "0.0.3"


class AttriDict(dict, AttriDictAttributes):

	'''AttriDict'''
	def __init__(self, data = {}, deep = True):
		verified_dict = self.__type_verification(data)
		if verified_dict != None:
			data = verified_dict
		else:
			raise AttributeError("argument not valid")

		super(type(self), self).__init__(data)

		self.__to_attridict(data, deep)

		self.__reset()

	def __call__(self, data = {}):
		self = AttriDict()
		verified_dict = self.__type_verification(data)

		if verified_dict != None:
			data = verified_dict
		else:
			raise AttributeError("argument not valid")

		super(AttriDict, self).__init__(data)

		self.__to_attridict(data, deep)

		self.__reset()
		return self

	def __contains__(self):
		return self.__dict__.__contains__(key)

	def __delitem__(self, key):
		self.__dict__.__delitem__(key)

	def __dir__(self):
		return dir(self.__class__) + list(self.__dict__.keys())

	def __eq__(self, other):
		return self.__dict__.__eq__(other)

	def __getitem__(self, key):
		return self.__dict__.__getitem__(key)

	def __iter__(self):
		return len(self.__dict__.keys())

	def __len__(self):
		return len(self.__dict__.keys())

	def __ne__(self, other):
		return self.__dict__.__ne__(other)

	def __reduce__(self):
		return self.__dict__.__reduce__()

	def __repr__(self):
		return str(self.__dict__)

	def __setattr__(self, key, value):
		if not self.__valid_key(key):
			raise AttributeError("invalid key, '{0}'".format(key))

		if type(value) == dict:
			value = type(self)(value)
		
		self.__dict__.__setitem__(key, value)

		self.__reset()

	def __setitem__(self, key, value):
		self.__dict__.__setitem__(key, value)

	def __str__(self):
		return str(self.__dict__)


	def __copy__(self):
		return self.__class__(self)

	def __to_string__(self):
		return f"{{@{str(self.__dict__)[1:-1]}@}}"		

	def __reset(self):
		'''reset dict'''
		dict.__init__(self, self.__dict__)
		# super(AttriDict, self).__init__(self.__dict__)


	def __to_attridict(self, data, deep):
		for key, value in data.items():
			if deep:
				typ = type(value)
				if typ == dict:
					self.__dict__[key] = type(self)(value, deep)

				elif typ in [list, tuple]:
					value = list(value)

					for i, val in enumerate(value):
						if type(val) == dict:
							value[i] = type(self)(val, deep)
					
					self.__dict__[key] = typ(value)
				else:
					self.__dict__[key] = value
			else:
				self.__dict__[key] = value

	def __to_dict(self):
		for key, value in self.__dict__.items():
			typ = type(value)
			if typ == type(self):
				self.__dict__[key] = value.to_dict()

			elif typ in [list, tuple]:
				value = list(value)

				for i, val in enumerate(value):
					if type(val) == type(self):
						value[i] = val.to_dict()

				self.__dict__[key] = typ(value)
			else:	
				self.__dict__[key] = value


	def __try_dict_convertion(self, wannabe_dict):
		try:
			return dict(wannabe_dict)
		except Exception as err:
			raise AttributeError(err)


	def __type_verification(self, data):
		if type(data) != dict or type(data) != AttriDict:
			return self.__try_dict_convertion(data)
		return data


	def __valid_key(self, word):
		if word in dir(dict):
			return False
		return True


	def to_dict(self):
		self.__to_dict()
		return self.__dict__



	def clear(self):
		'''D.clear() -> None. Remove all items from D'''
		self.__dict__.clear()

	def copy(self):
		'''D.copy() -> a shallow copy of D'''
		# return self.__class__(self)
		return self.__copy__()

	def get(self, key, default = None):
		'''Return the value for key if key is in the attribute dictionary, else default.'''
		return self.__dict__.get(key, default)

	def items(self, *args, **kwargs):
		'''D.items() -> a set-like object providing a view on D's items'''
		return self.__dict__.items(*args, **kwargs)

	def keys(self):
		'''D.keys() -> a set-like object providing a view on D's keys'''
		return self.__dict__.keys()

	def pop(self, key, value = None):
		'''D.pop(k[,d]) -> v, remove specified key and return the corresponding value.\
		\nIf key is not found, d is returned if given, otherwise KeyError is raised
		'''
		return self.__dict__.pop(key, value)

	def popitem(self):
		return self.__dict__.popitem()

	def setdefault(self, key, default=None):
		'''Insert key with a value of default if key is not in the dictionary.\
		\n\nReturn the value for key if key is in the dictionary, else default.'''
		return self.__dict__.setdefault(key, default)

	def update(self, *args, **kwargs):
		return self.__dict__.update(*args, **kwargs)

	def values(self):
		'''D.values() -> an object providing a view on D's values'''
		return self.__dict__.values()


if __name__ == "attridict":
	import sys
	sys.modules[__name__] = AttriDict



#!/usr/bin/env python3

'''
python -m unittest tests.py
'''

import unittest

import attridict


class TestAttriDict(unittest.TestCase):
	def test_attridict(self):
		att = attridict()
		self.assertEqual(att, {})

		att.one = 111
		self.assertEqual(att.one, 111)
		self.assertEqual(att, {"one": 111})

		att.two = {"three": 333}
		self.assertEqual(att.two, {"three": 333})
		self.assertEqual(att, {"one": 111, "two": {"three": 333}})

		att.two.four = {"five": 555, "six": 666}
		self.assertEqual(att.two.four.five, 555)
		self.assertEqual(att.two.four.six, 666)

		self.assertEqual(att, {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}})


	def test_attridict2(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		self.assertEqual(att.one, 111)
		self.assertEqual(att.two, {"three": 333, "four": {"five": 555, "six": 666}})
		self.assertEqual(att.two.three, 333)
		self.assertEqual(att.two.four, {"five": 555, "six": 666})
		self.assertEqual(att.two.four.five, 555)
		self.assertEqual(att.two.four.six, 666)

		att.two = 222
		self.assertEqual(att.two, 222)

		att.three = {"four": 444, "five": 555}
		self.assertEqual(att.three, {"four": 444, "five": 555})
		self.assertEqual(att, {"one": 111, "two": 222, "three": {"four": 444, "five": 555}})


	def test_attridict3(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		self.assertEqual(att, data)
		self.assertEqual(att.one, data["one"])
		self.assertEqual(att.two, data["two"])
		self.assertEqual(att.two.three, data["two"]["three"])
		self.assertEqual(att.two.four, data["two"]["four"])
		self.assertEqual(att.two.four.five, data["two"]["four"]["five"])
		self.assertEqual(att.two.four.six, data["two"]['four']["six"])


	def test_dir(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		for i in data.keys():
			self.assertIn(i, att.__dir__())

		for i in data["two"].keys():
			self.assertIn(i, att.two.__dir__())

		for i in data["two"]["four"].keys():
			self.assertIn(i, att.two.four.__dir__())


	def test_attridict5(self):
		# #############################################################################
		# # problem tests
		# #############################################################################

		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)
		att.three = {"four": 444, "five": 565}

		del att

		# #############################################################################

		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		self.assertEqual(data.__len__(), att.__len__())
		self.assertEqual(data["two"].__len__(), att.two.__len__())
		self.assertEqual(data["two"]["four"].__len__(), att.two.four.__len__())

		# ################################################################################


	def test_attridict6(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)
		self.assertEqual(att, data)


	def test_mutable(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)
		att_mutable = att
		att.two = 222
		self.assertEqual(att, att_mutable)
		self.assertIs(att, att_mutable)


	def test_copy(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)
		att_copy = att.copy()
		
		self.assertIsNot(att, att_copy)

		att.two = 222
		self.assertNotEqual(att, att_copy)
		self.assertEqual(att_copy, data)


	def test_to_dict(self):
		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		att_dict = att.to_dict()
		self.assertEqual(data, att_dict)

		self.assertEqual(type(att_dict), dict)
		self.assertEqual(type(att_dict["two"]), dict)
		self.assertEqual(type(att_dict["two"]["four"]), dict)


	def test_attridict10(self):

		data = {"one": 111, "two": {"three": 333, "four": {"five": 555, "six": 666}}}
		att = attridict(data)

		del att.two.four.six
		self.assertEqual(att, {"one": 111, "two": {"three": 333, "four": {"five": 555}}})


	def test_deep_true(self):
		data = {"one": 111, "two": [{"three": 333, "four": 444}]}
		att = attridict(data, True)
		
		self.assertEqual(type(att.two[0]), attridict)
		self.assertEqual(att.two[0], {"three": 333, "four": 444})
		self.assertEqual(att.two[0].three, 333)


	def test_deep_false(self):
		data = {"one": 111, "two": [{"three": 333, "four": 444}]}
		att = attridict(data, False)

		self.assertEqual(type(att.two[0]), dict)
		self.assertEqual(att.two[0], {"three": 333, "four": 444})

	def test_to_dict_deep_list(self):
		data = {"one": 111, "two": [{"three": 333, "four": 444}]}
		att = attridict(data, True)
	
		self.assertEqual(type(att.two[0]), attridict)
		self.assertEqual(att.two[0], {"three": 333, "four": 444})
		self.assertEqual(att.two[0].three, 333)

		d_attr = att.to_dict()

		with self.assertRaises(AttributeError) as ctx:
			d_attr["two"][0].three
		self.assertEqual(str(ctx.exception), "'dict' object has no attribute 'three'")


	def test_to_dict_deep_tuple(self):
		data = {"one": 111, "two": tuple(({"three": 333, "four": 444},))}
		att = attridict(data, True)
	
		self.assertEqual(type(att.two[0]), attridict)
		self.assertEqual(att.two[0], {"three": 333, "four": 444})
		self.assertEqual(att.two[0].three, 333)

		d_attr = att.to_dict()

		with self.assertRaises(AttributeError) as ctx:
			d_attr["two"][0].three
		self.assertEqual(str(ctx.exception), "'dict' object has no attribute 'three'")

	def test_to_string_dunder_method(self):
		data = {"one": 111, "two": 222}
		att = attridict(data)

		self.assertEqual(att.__to_string__(), "{@'one': 111, 'two': 222@}")

if __name__ == "__main__":
	unittest.main()



'''
python setup.py install
'''

from setuptools import setup

import attridict

readme = open("README.md", 'r')
long_description = readme.read()

setup(
	name="attridict",
	version=attridict.__version__,
	author="Alvin Shaita",
	author_email="alvinshaita@gmail.com",
	url="https://github.com/alvinshaita/attridict",
	license="MIT License",
	packages=["."],
	keywords="attridict, attrdict, struct, dict, dot, attribute, attributes, dictionary, attr",
	description="A dict implementation with support for easy and clean access of its values through attributes",
	long_description=long_description,
	long_description_content_type='text/markdown',
	classifiers=[
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 2",
        	"Programming Language :: Python :: 3",
        	"Programming Language :: Python :: Implementation :: CPython",
        	"Programming Language :: Python :: Implementation :: PyPy",
	]
)
